from logging import debug, error, info
from os import environ
from cgi import parse_qs
from google.appengine.api.memcache import get, set as mset, get_multi, delete as mdel
from random import sample

"""
A ntrack tracker
================

Memcached namespaces:

- 'K': Keys / info_hashes -> set of ips (sets have a slight overhead over lists, but are more foolsafe)
- 'I': Client ips -> metadata (currently a tuple with one item: (Port,)) XXX the same client could be in multiple tracks with different ips! (Think NAT!)
- 'D': Debug data

This allows peer info to be shared and decay by itself, we will delete references to peer from
the key namespace lazily.
"""

def resps(s):
    print "Content-type: text/plain"
    print ""
    print s, # Make sure we don't add a trailing new line!

def key2s(k):
    if k:
        return ''.join(["%03X"% ord(c) for c in k])
    else:
        return "NO KEY!"


def prof_main():
    # This is the main function for profiling 
    # We've renamed our original main() above to real_main()
    import cProfile, pstats, StringIO
    import logging
    prof = cProfile.Profile()
    prof = prof.runctx("real_main()", globals(), locals())
    stream = StringIO.StringIO()
    stats = pstats.Stats(prof, stream=stream)
    stats.sort_stats("time")  # Or cumulative
    stats.print_stats(80)  # 80 = how many to print
    # The rest is optional.
    # stats.print_callees()
    # stats.print_callers()
    logging.info("Profile data:\n%s", stream.getvalue())


def real_main():
    args = parse_qs(environ['QUERY_STRING'])

    if not args:
        print "Status: 301 Moved Permanantly"
        print "Location: /"
        print ""
        return

    for a in ['info_hash', 'port']:
        if a not in args or len(args[a]) < 1:
            return # We get many of this, don't waste bw or cpu on them!
            info("Missing required argument: %s ..."%a)
            resps(bencode({'failure reason': "You must provide %s!"%a}))

    key = args['info_hash'][0]
    if(len(key) > 128):
        info("Rejected too big key: %s ..."%repr(key[:128]))
        resps(bencode({'failure reason': "Insanely long key!"}))
        return
    ip = environ['REMOTE_ADDR']
    port = int(args['port'][0]) # TODO Should catch str->int conversion errors
    # TODO BT: If left=0, the download is done and we should not return any peers.
    event = args.pop('event', [None])[0]
    if event == "stopped":
        # XXX We should only remove it from this track, but this is good enough.
        mdel(ip, namespace='I')
        return # They are going away, don't waste bw/cpu on this.
        debug("Deleting peer %s"%ip)
        resps(bencode({'interval': 2048, 'peers': []}))

    updatetrack = False

    # Get existing peers
    s = get(key, namespace='K')

    if s:
        # TODO rate limiting, exponential backoff, etc

        if len(s) > 32:
            ips = set(sample(s, 32))
        else:
            ips = s

        peers = get_multi(ips, namespace='I')

        lostpeers = [p for p in ips if p not in peers]
        if lostpeers: # Remove lost peers
            #info("Removed 'lost' peers: "+repr(lostpeers)+" from track: "+key2s(key)) 
            s.difference_update(lostpeers)
            updatetrack = True

        peers.pop(ip, None) # Remove self from list of returned peers

    # New track! Create track with this ip and we are done!
    else:
        s = set([])
        peers = {}

    mset(ip, (port,), namespace='I') # This might be redundant, but ensures we update the port number in case it has changed.
    if ip not in s: # Assume new peer
        s.add(ip)
        updatetrack = True

    if updatetrack: 
        mset(key, s, namespace='K')

    #debug("Returned %s peers" % len(peers))
    resps(bencode({'interval': 4424, 'peers': [{'ip': p, 'port': peers[p][0]} for p in peers]}))

main = prof_main
#main = real_main

################################################################################
# Bencode encoding code by Petru Paler, slightly simplified by uriel
from types import StringType, IntType, LongType, DictType, ListType, TupleType, BooleanType

#class Bencached(object):
#    __slots__ = ['bencoded']
#
#    def __init__(self, s):
#        self.bencoded = s

#def encode_bencached(x,r):
#    r.append(x.bencoded)

def encode_int(x, r):
    r.extend(('i', str(x), 'e'))

def encode_string(x, r):
    r.extend((str(len(x)), ':', x))

def encode_list(x, r):
    r.append('l')
    for i in x:
        encode_func[type(i)](i, r)
    r.append('e')

def encode_dict(x,r):
    r.append('d')
    ilist = x.items()
    ilist.sort()
    for k, v in ilist:
        r.extend((str(len(k)), ':', k))
        encode_func[type(v)](v, r)
    r.append('e')

encode_func = {}
#encode_func[type(Bencached(0))] = encode_bencached
encode_func[IntType] = encode_int
encode_func[LongType] = encode_int
encode_func[StringType] = encode_string
encode_func[ListType] = encode_list
encode_func[TupleType] = encode_list
encode_func[DictType] = encode_dict
encode_func[BooleanType] = encode_int

def bencode(x):
    r = []
    encode_func[type(x)](x, r)
    return ''.join(r)

def test_bencode():
    assert bencode(4) == 'i4e'
    assert bencode(0) == 'i0e'
    assert bencode(-10) == 'i-10e'
    assert bencode(12345678901234567890L) == 'i12345678901234567890e'
    assert bencode('') == '0:'
    assert bencode('abc') == '3:abc'
    assert bencode('1234567890') == '10:1234567890'
    assert bencode([]) == 'le'
    assert bencode([1, 2, 3]) == 'li1ei2ei3ee'
    assert bencode([['Alice', 'Bob'], [2, 3]]) == 'll5:Alice3:Bobeli2ei3eee'
    assert bencode({}) == 'de'
    assert bencode({'age': 25, 'eyes': 'blue'}) == 'd3:agei25e4:eyes4:bluee'
    assert bencode({'spam.mp3': {'author': 'Alice', 'length': 100000}}) == 'd8:spam.mp3d6:author5:Alice6:lengthi100000eee'
    #assert bencode(Bencached(bencode(3))) == 'i3e'
    try:
        bencode({1: 'foo'})
    except TypeError:
        return
    assert 0

################################################################################

if __name__ == '__main__':
    main()
